/*
 *	undll.exe Ver.2.04	Copyright (C) 1999-2011  K.Takata
 */

#define WIN32_LEAN_AND_MEAN
#define NOGDI
#include <windows.h>

#ifndef lengthof
#define lengthof(arr)	(sizeof(arr) / sizeof((arr)[0]))
#endif

//#define BUFFSIZE (1536 * 1024)			// 1.5MiB
#define BUFFSIZE (4 * 1024 * 1024)			// 4MiB
#define HELPSIZE (256 * 1024)				// 256KiB

typedef int (WINAPI *ARC_MAINPROC)(const HWND, LPCSTR, LPSTR, const DWORD);
typedef WORD (WINAPI *ARC_GETVERSION)(VOID);


HANDLE g_hStdOut;			// handle for standard output
char *g_buf;


#define CHUNKSIZE	0x8000
void myfputs(const char *str, HANDLE handle)
{
	DWORD dwWritten;
	DWORD len, m, n, i;
	
	len = lstrlen(str);
	m = len / CHUNKSIZE;
	n = len & (CHUNKSIZE - 1);
	
	for (i = 0; i <= m; i++) {
		WriteFile(handle, str, (i == m) ? n : CHUNKSIZE, &dwWritten, NULL);
		str += CHUNKSIZE;
	}
}

void putmsg(const char *str)
{
	myfputs(str, g_hStdOut);
}

void copyright(void)
{
	char *msg =
		"undll.exe Ver.2.04   Copyright (C) 1999-2011  K.Takata"
		"\r\n\n";
	
	putmsg(msg);
}

DWORD puterrmsg(void)
{
	char buf[260];
	DWORD err;
	
	err = GetLastError();
	FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM
			| FORMAT_MESSAGE_ARGUMENT_ARRAY,
			NULL, err, LANG_NEUTRAL, buf, lengthof(buf), NULL);
	
	putmsg(buf);
	return err;
}

void help(const char *helpfile)
{
	HANDLE hFile;
	DWORD size;
	
	if (helpfile == NULL || helpfile[0] == '\0') {
		return;
	}
	hFile = CreateFile(helpfile, GENERIC_READ, FILE_SHARE_READ, NULL,
			OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN, NULL);
	if (hFile == INVALID_HANDLE_VALUE) {
	//	putmsg("Can't open the helpfile.\n");
		puterrmsg();
		return;
	}
	// Read Helpfile and Print it
	if (ReadFile(hFile, g_buf, HELPSIZE, &size, NULL)) {
		if (size >= HELPSIZE) {
			size = HELPSIZE;
		}
		g_buf[size] = '\0';
		putmsg(g_buf);
	} else {
	//	putmsg("Can't read.\n");
		puterrmsg();
	}
	
	CloseHandle(hFile);
}


int main()
{
	char path[MAX_PATH];
	char sectname[MAX_PATH];
	char name[MAX_PATH];
	char *func = (char *) 1;		// ordinal 1;
//	char buf[BUFFSIZE];
	int ret, buffsize;
	unsigned char *lpszCommandLine, *p;//, *fname, *lpszOrgCommandLine;
//	WORD wVer;
	ARC_MAINPROC ArcMainProc;
//	ARC_GETVERSION ArcGetVersion;
	HMODULE hDLL = NULL;
	
	
	// Get handle for standard output
	g_hStdOut = GetStdHandle(STD_OUTPUT_HANDLE);
	
	
//	fname = lpszCommandLine = lpszOrgCommandLine = GetCommandLine();
	lpszCommandLine = GetCommandLine();
	
	// Skip past program name (first token in command line)
	
	// Check for and handle quoted program name
	if (*lpszCommandLine == '"') {
		// Scan, and skip over, subsequent characters until another
		// double-quote or a null is encountered
		do {
			lpszCommandLine++;
		} while (*lpszCommandLine && (*lpszCommandLine != '"'));
		
		// If we stopped on a double-quote (usual case), skip over it.
		if (*lpszCommandLine == '"') {
			*lpszCommandLine++ = '\0';
		}
	} else {
		// First token wasn't a quote
		while (*lpszCommandLine > ' ') {
			lpszCommandLine++;
		}
	}
	
	// skip past any white space preceeding the second token.
	while (*lpszCommandLine && (*lpszCommandLine <= ' ')) {
		*lpszCommandLine++ = '\0';
	}
	
	
	
	GetModuleFileName(NULL, path, lengthof(path));
	
	// Get Base File Name
	p = path + lstrlen(path);
	if (*(p - 4) == '.') {		// cut ".ext" (maybe ".exe")
		p -= 4;
		*p = '\0';
	}
	while (--p > path) {		// BUG (NOT support Japanese exec file name)
		if (*p == '\\') {
			++p;
			break;
		}
	}
	lstrcpy(name, p);			// base file name
	
	// Set .ini File Name
	((DWORD*) p)[0] = 0x6c646e75;		// lstrcpy(p, "undll.ini");
	((DWORD*) p)[1] = 0x6e692e6c;
	(( WORD*) p)[4] = 0x0069;
	
	
	// Get Actual Section Name
	GetPrivateProfileString(name, "ref", name,
			sectname, lengthof(sectname), path);
	
	
	// Get Buffer Size
	buffsize = GetPrivateProfileInt(sectname, "bufsiz", BUFFSIZE, path);
	if (buffsize < HELPSIZE)
		buffsize = HELPSIZE;
	
	// Allocate a Buffer
	g_buf = VirtualAlloc(NULL, buffsize, MEM_COMMIT, PAGE_READWRITE);
	if (g_buf == NULL) {
	//	putmsg("Not enough memory.\n");
		puterrmsg();
		ret = 4;
		goto end;
	}
	
	// Display Help
	if (*lpszCommandLine == '\0') {		// no args
		char *helpfile = name;
		copyright();
		// Get Helpfile Name
		GetPrivateProfileString(sectname, "help", "",
				name, lengthof(name), path);
		if (name[0] == '>') {	// relative to the module path
			lstrcpy(p, name + 1);
			helpfile = path;
		}
		help(helpfile);
		ret = 1;
		goto end;
	}
	
	// Get DLL Name
	GetPrivateProfileString(sectname, "DLL", sectname,
			name, lengthof(name), path);
	
	hDLL = LoadLibrary(name);
	if (hDLL == NULL) {
	//	putmsg("Can't load the DLL.\n");
		puterrmsg();
		ret = 2;
		goto end;
	}
	
	// Get Main Proc Name
	GetPrivateProfileString(sectname, "Proc", "",
			name, lengthof(name), path);
	
	if (name[0] != '\0') {
		func = name;
	}
	ArcMainProc = (ARC_MAINPROC) GetProcAddress(hDLL, func);
	if (ArcMainProc == NULL) {
	//	putmsg("Can't get Proc Address.\n");
		puterrmsg();
		ret = 3;
		goto end;
	}
	// Execute Archiver DLL
	ret = ArcMainProc(NULL, lpszCommandLine, g_buf, buffsize);
	
	// Print Result
	g_buf[buffsize - 1] = '\0';
	putmsg(g_buf);
	
end:
	/*
	 * These resources will be released automatically
	 * when this program ends.
	 */
/*
	if (hDLL != NULL) {
		FreeLibrary(hDLL);
	}
	if (g_buf != NULL) {
		VirtualFree(g_buf, 0, MEM_RELEASE | MEM_DECOMMIT);
	}
*/
	return ret;
}
